(******************************************************************************)
(* generator (stack machine)                                                  *)
(* (C) 2010..2021 Steffen Solyga <solyga@gmx.de>                              *)
(******************************************************************************)
IMPLEMENTATION MODULE Gen;

FROM SYSTEM IMPORT ADR, BYTE;
IMPORT Lib, FIO;
IMPORT Comp0;
FROM Comp IMPORT Write, GetPos, SetPos, Read;

CONST
  debug = TRUE;
  err ::= FIO.ErrorOutput;
  dbg ::= FIO.ErrorOutput;

VAR
  (* debug *)
  first: BOOLEAN;
  (* constants *)
  op0s: ARRAY TOp0 OF Comp0.TOperation;
  op1s: ARRAY TOp1 OF Comp0.TOperation;
  op2s: ARRAY TOp2 OF Comp0.TOperation;
  xops: ARRAY TOp2 OF Comp0.TOperation;
  vals: ARRAY BOOLEAN OF TValue;

PROCEDURE DbgC( cmd: Comp0.TCmdDesc );
VAR a: TAddress; s: Comp0.TCmdStr;
BEGIN
  GetPos( a );
  Comp0.Convert( cmd, s );
  IF first THEN FIO.WrStr( dbg, 'Gen.Dbg:  pos' ); FIO.WrLn( dbg ) END;
  first := FALSE;
  FIO.WrStr( dbg, 'Gen.Dbg: ' );
  FIO.WrCard( dbg, a, 4 );
  FIO.WrStr( dbg, ': ' );
  FIO.WrStr( dbg, s );
  FIO.WrLn( dbg );
END DbgC;

PROCEDURE DbgF( a: TAddress );
BEGIN
  FIO.WrStr( dbg, 'Gen.Dbg: Fixup ' );
  FIO.WrCard( dbg, a, 0 );
  FIO.WrLn( dbg );
END DbgF;

PROCEDURE Err( s: ARRAY OF CHAR );
BEGIN
  FIO.WrStr( err, 'Gen: ' );
  FIO.WrStr( err, s );
  FIO.WrStr( err, '.' );
  FIO.WrLn( err );
  HALT;
END Err;


PROCEDURE Hlt();
VAR cmd: Comp0.TCmdDesc;
BEGIN
  cmd.fct := Comp0.fHlt;
  IF debug THEN DbgC( cmd ) END;
  Write( cmd.fct );
END Hlt;

PROCEDURE Val( v: TValue );
VAR cmd: Comp0.TCmdDesc;
BEGIN
  cmd.fct := Comp0.fVal; cmd.val := v;
  IF debug THEN DbgC( cmd ) END;
  Write( cmd.fct ); Write( cmd.val );
END Val;

PROCEDURE Sto( i: TIndex );
VAR cmd: Comp0.TCmdDesc;
BEGIN
  IF i >= Comp0.vcMax THEN Err( 'Invalid index' ) END;
  cmd.fct := Comp0.fSto; cmd.idx := i;
  IF debug THEN DbgC( cmd ) END;
  Write( cmd.fct ); Write( cmd.idx );
END Sto;

PROCEDURE Lod( i: TIndex );
VAR cmd: Comp0.TCmdDesc;
BEGIN
  IF i >= Comp0.vcMax THEN Err( 'Invalid index' ) END;
  cmd.fct := Comp0.fLod; cmd.idx := i;
  IF debug THEN DbgC( cmd ) END;
  Write( cmd.fct ); Write( cmd.idx );
END Lod;

PROCEDURE Opr( o: Comp0.TOperation );
VAR cmd: Comp0.TCmdDesc;
BEGIN
  IF o = Comp0.oNull THEN Err( 'Bug: Invalid operation' ) END;
  cmd.fct := Comp0.fOpr; cmd.opr := o;
  IF debug THEN DbgC( cmd ) END;
  Write( cmd.fct ); Write( cmd.opr );
END Opr;


PROCEDURE Op0( o: TOp0 );
BEGIN
  Opr( op0s[o] );
END Op0;

PROCEDURE Op1( o: TOp1; VAR d: TItemDesc );
BEGIN
  IF d.c THEN
    CASE o OF
    | oNeg: d.v := -d.v;
    | oNot: d.v := TValue( Comp0.TSet(d.v) / Comp0.TSet(Comp0.TValue(-1)) );
    ELSE
      Err( 'Bug: Unknown unary operator' );
    END;
  ELSE
    Opr( op1s[o] );
  END;
END Op1;

PROCEDURE Op2( o: TOp2; VAR d, d1: TItemDesc );
BEGIN
  IF d.c & d1.c THEN
    CASE o OF
    | oAdd: d.v := d.v + d1.v;
    | oSub: d.v := d.v - d1.v;
    | oMul: d.v := d.v  *  d1.v;
    | oDiv: d.v := d.v DIV d1.v;
    | oOr : d.v := TValue( Comp0.TSet(d.v) + Comp0.TSet(d1.v) );
    | oAnd: d.v := TValue( Comp0.TSet(d.v) * Comp0.TSet(d1.v) );
    | oEql: d.v := vals[ d.v =  d1.v ];
    | oNeq: d.v := vals[ d.v #  d1.v ];
    | oLss: d.v := vals[ d.v <  d1.v ];
    | oLeq: d.v := vals[ d.v <= d1.v ];
    | oGtr: d.v := vals[ d.v >  d1.v ];
    | oGeq: d.v := vals[ d.v >= d1.v ];
    ELSE
      Err( 'Bug: Unknown binary operator' );
    END;
  ELSE
    IF d.c THEN
      Val( d.v );
      Opr( xops[o] );
    ELSE
      IF d1.c THEN Val( d1.v ) END;
      Opr( op2s[o] );
    END;
    d.c := FALSE;
  END;
END Op2;

PROCEDURE Label( VAR a: TAddress );
BEGIN
  GetPos( a );
END Label;

PROCEDURE Jmp( a: TAddress );
VAR cmd: Comp0.TCmdDesc; l: TAddress;
BEGIN
  GetPos( l ); IF a > l THEN Err( 'Bug: Invalid address' ) END;
  cmd.fct := Comp0.fJmp; cmd.adr := a;
  IF debug THEN DbgC( cmd ) END;
  Write( cmd.fct ); Write( cmd.adr );
END Jmp;

PROCEDURE Jpf( a: TAddress; d: TItemDesc );
VAR cmd: Comp0.TCmdDesc; l: TAddress;
BEGIN
  GetPos( l ); IF a > l THEN Err( 'Bug: Invalid address' ) END;
  IF d.c THEN Val( d.v ) END;
  cmd.fct := Comp0.fJpz; cmd.adr := a;
  IF debug THEN DbgC( cmd ) END;
  Write( cmd.fct ); Write( cmd.adr );
END Jpf;

PROCEDURE Init( VAR a: TAddress );
BEGIN
  a := Comp0.aNone;
END Init;

PROCEDURE JmpFwd( VAR a: TAddress );
VAR cmd: Comp0.TCmdDesc;
BEGIN
  cmd.fct := Comp0.fJmp; cmd.adr := a;
  IF debug THEN DbgC( cmd ) END;
  Write( cmd.fct ); GetPos( a ); Write( cmd.adr );
END JmpFwd;

PROCEDURE JpfFwd( VAR a: TAddress; d: TItemDesc );
VAR cmd: Comp0.TCmdDesc;
BEGIN
  IF d.c THEN Val( d.v ) END;
  cmd.fct := Comp0.fJpz; cmd.adr := a;
  IF debug THEN DbgC( cmd ) END;
  Write( cmd.fct ); GetPos( a ); Write( cmd.adr );
END JpfFwd;

PROCEDURE Fixup( a: TAddress );
VAR pos, nxt: TAddress;
BEGIN
  GetPos( pos );
  WHILE a # Comp0.aNone DO
    IF debug THEN DbgF( a ) END;
    SetPos( a ); Read( nxt );
    SetPos( a ); Write( pos );
    a := nxt;
  END;
  SetPos( pos );
END Fixup;


PROCEDURE InitModule();
BEGIN
  first := TRUE;

  Lib.Fill( ADR(op0s), SIZE(op0s), 0C );
  op0s[ oWrt ] := Comp0.oWrt;
  op0s[ oWrB ] := Comp0.oWrB;

  Lib.Fill( ADR(op1s), SIZE(op1s), 0C );
  op1s[ oNeg ] := Comp0.oNeg;
  op1s[ oNot ] := Comp0.oNot;

  Lib.Fill( ADR(op2s), SIZE(op2s), 0C );
  op2s[ oAdd ] := Comp0.oAdd;
  op2s[ oSub ] := Comp0.oSub;
  op2s[ oMul ] := Comp0.oMul;
  op2s[ oDiv ] := Comp0.oDiv;
  op2s[ oOr  ] := Comp0.oOr;
  op2s[ oAnd ] := Comp0.oAnd;
  op2s[ oEql ] := Comp0.oEql;
  op2s[ oNeq ] := Comp0.oNeq;
  op2s[ oLss ] := Comp0.oLss;
  op2s[ oLeq ] := Comp0.oLeq;
  op2s[ oGtr ] := Comp0.oGtr;
  op2s[ oGeq ] := Comp0.oGeq;

  Lib.Fill( ADR(xops), SIZE(xops), 0C );
  xops[ oAdd ] := Comp0.oAdd;
  xops[ oSub ] := Comp0.oBus;
  xops[ oMul ] := Comp0.oMul;
  xops[ oDiv ] := Comp0.oVid;
  xops[ oOr  ] := Comp0.oOr;
  xops[ oAnd ] := Comp0.oAnd;
  xops[ oEql ] := Comp0.oEql;
  xops[ oNeq ] := Comp0.oNeq;
  xops[ oLss ] := Comp0.oGtr;
  xops[ oLeq ] := Comp0.oGeq;
  xops[ oGtr ] := Comp0.oLss;
  xops[ oGeq ] := Comp0.oLeq;

  Lib.Fill( ADR(vals), SIZE(vals), 0C );
  vals[ FALSE ] := Comp0.vFalse;
  vals[ TRUE  ] := Comp0.vTrue;
END InitModule;

BEGIN
  InitModule();
END Gen.
